import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
from patsy import dmatrices, dmatrix, demo_data

from .. import families as fam
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
from .. import gas as gas
from .. import data_check as dc

class EGARCHMReg(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** BETA-t-EGARCH IN MEAN REGRESSION MODELS ****

    Parameters
    ----------
    data : pd.DataFrame
        Field to specify the time series data that will be used.

    p : int
        Field to specify how many GARCH terms the model will have.

    q : int
        Field to specify how many SCORE terms the model will have.

    formula : string
        patsy string describing the regression

    """

    def __init__(self, data, p, q, formula):

        # Initialize TSM object
        super(EGARCHMReg,self).__init__('EGARCHMReg')

        # Latent variables
        self.p = p
        self.q = q
        self.max_lag = max(self.p,self.q)  
        self.z_no = self.p + self.q + 2
        self._z_hide = 0 # Whether to cutoff variance latent variables from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"
        self.multivariate_model = False
        self.leverage = False
        self.model_name = "EGARCHMReg(" + str(self.p) + "," + str(self.q) + ")"

        # Format the data
        self.is_pandas = True # This is compulsory for this model type
        self.data_original = data
        self.formula = formula
        self.y, self.X = dmatrices(formula, data)
        self.z_no += self.X.shape[1]*2
        self.y_name = self.y.design_info.describe()
        self.data_name = self.y_name
        self.X_names = self.X.design_info.describe().split(" + ")
        self.y = np.array([self.y]).ravel()
        self.data = self.y
        self.data_length = len(self.data)
        self.X = np.array([self.X])[0]
        self.index = data.index
        self.initial_values = np.zeros(self.z_no)

        self._create_latent_variables()

    def _create_latent_variables(self):
        """ Creates model latent variables

        Returns
        ----------
        None (changes model attributes)
        """

        for p_term in range(self.p):
            self.latent_variables.add_z('p(' + str(p_term+1) + ')', fam.Normal(0,0.5,transform='logit'), fam.Normal(0,3))
            if p_term == 0:
                self.latent_variables.z_list[-1].start = 3.00
            else:
                self.latent_variables.z_list[-1].start = -4.00

        for q_term in range(self.q):
            self.latent_variables.add_z('q(' + str(q_term+1) + ')', fam.Normal(0,0.5,transform='logit'), fam.Normal(0,3))
            if q_term == 0:
                self.latent_variables.z_list[-1].start = -1.50  
            else: 
                self.latent_variables.z_list[-1].start = -4.00  

        self.latent_variables.add_z('v', fam.Flat(transform='exp'), fam.Normal(0,3))
        self.latent_variables.add_z('GARCH-M', fam.Normal(0, 3, transform=None),fam.Normal(0,3))

        for parm in range(len(self.X_names)):
            self.latent_variables.add_z('Vol Beta ' + self.X_names[parm], fam.Normal(0,10,transform=None), fam.Normal(0,3))

        for parm in range(len(self.X_names)):
            self.latent_variables.add_z('Returns Beta ' + self.X_names[parm], fam.Normal(0,10,transform=None), fam.Normal(0,3))

        # Starting values        

        for i in range(self.p+self.q, self.z_no):
            self.latent_variables.z_list[i].start = 0.0

        self.latent_variables.z_list[self.p+self.q].start = 2.0

    def _model(self, beta):
        """ Creates the structure of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        lambda : np.array
            Contains the values for the conditional volatility series

        Y : np.array
            Contains the length-adjusted time series (accounting for lags)

        scores : np.array
            Contains the score terms for the time series
        """

        Y = np.array(self.data[self.max_lag:self.data.shape[0]])
        X = np.ones(Y.shape[0])
        scores = np.zeros(Y.shape[0])

        # Transform latent variables
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])

        lmda = np.zeros(Y.shape[0])
        theta = np.zeros(Y.shape[0])

        # Loop over time series
        for t in range(0,Y.shape[0]):
            if t < self.max_lag:
                lmda[t] = parm[-len(self.X_names)*2]/(1-np.sum(parm[:self.p]))
                theta[t] = np.dot(self.X[t],parm[-len(self.X_names):])
            else:
                # Loop over GARCH terms
                for p_term in range(0,self.p):
                    lmda[t] += parm[p_term]*lmda[t-p_term-1]

                # Loop over Score terms
                for q_term in range(0,self.q):
                    lmda[t] += parm[self.p+q_term]*scores[t-q_term-1]

                if self.leverage is True:
                    lmda[t] += parm[-(len(self.X_names)*2)-3]*np.sign(-(Y[t-1]-theta[t-1]))*(scores[t-1]+1)

                lmda[t] += np.dot(self.X[t],parm[-len(self.X_names)*2:-len(self.X_names)])

                theta[t] = np.dot(self.X[t],parm[-len(self.X_names):]) + parm[-(len(self.X_names)*2)-1]*np.exp(lmda[t]/2.0)
            
            scores[t] = (((parm[self.p+self.q]+1.0)*np.power(Y[t]-theta[t],2))/float(parm[self.p+self.q]*np.exp(lmda[t]) + np.power(Y[t]-theta[t],2))) - 1.0        

        return lmda, Y, scores, theta

    def _mb_model(self, beta, mini_batch):
        """ Creates the structure of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        lambda : np.array
            Contains the values for the conditional volatility series

        Y : np.array
            Contains the length-adjusted time series (accounting for lags)

        scores : np.array
            Contains the score terms for the time series
        """

        # Transform latent variables
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])

        rand_int =  np.random.randint(low=0, high=self.data_length-mini_batch+1)
        sample = np.arange(start=rand_int, stop=rand_int+mini_batch)

        data = self.y[sample]
        X = self.X[sample, :]
        Y = data[self.max_lag:]

        scores = np.zeros(Y.shape[0])
        lmda = np.ones(Y.shape[0])
        theta = np.ones(Y.shape[0])

        # Loop over time series
        for t in range(0,Y.shape[0]):
            if t < self.max_lag:
                lmda[t] = parm[-len(self.X_names)*2]/(1-np.sum(parm[:self.p]))
                theta[t] = np.dot(self.X[t],parm[-len(self.X_names):])
            else:
                # Loop over GARCH terms
                for p_term in range(0,self.p):
                    lmda[t] += parm[p_term]*lmda[t-p_term-1]

                # Loop over Score terms
                for q_term in range(0,self.q):
                    lmda[t] += parm[self.p+q_term]*scores[t-q_term-1]

                if self.leverage is True:
                    lmda[t] += parm[-(len(self.X_names)*2)-3]*np.sign(-(Y[t-1]-theta[t-1]))*(scores[t-1]+1)

                lmda[t] += np.dot(self.X[t],parm[-len(self.X_names)*2:-len(self.X_names)])

                theta[t] = np.dot(self.X[t],parm[-len(self.X_names):]) + parm[-(len(self.X_names)*2)-1]*np.exp(lmda[t]/2.0)
            
            scores[t] = (((parm[self.p+self.q]+1.0)*np.power(Y[t]-theta[t],2))/float(parm[self.p+self.q]*np.exp(lmda[t]) + np.power(Y[t]-theta[t],2))) - 1.0        

        return lmda, Y, scores, theta

    def _mean_prediction(self, lmda, Y, scores, h, t_params, X_oos):
        """ Creates a h-step ahead mean prediction

        Parameters
        ----------
        lmda : np.array
            The past predicted values

        Y : np.array
            The past data

        scores : np.array
            The past scores

        h : int
            How many steps ahead for the prediction

        t_params : np.array
            A vector of (transformed) latent variables

        X_oos : np.array
            Out of sample predictors

        Returns
        ----------
        h-length vector of mean predictions
        """     

        # Create arrays to iteratre over
        lmda_exp = lmda.copy()
        scores_exp = scores.copy()
        Y_exp = Y.copy()

        # Loop over h time periods          
        for t in range(0,h):
            new_lambda_value = 0

            if self.p != 0:
                for j in range(self.p):
                    new_lambda_value += t_params[j]*lmda_exp[-j-1]

            if self.q != 0:
                for k in range(self.q):
                    new_lambda_value += t_params[k+self.p]*scores_exp[-k-1]

            # No leverage term for mean (should be zero in expectation?)

            new_lambda_value += np.dot(X_oos[t],t_params[-len(self.X_names)*2:-len(self.X_names)]) 
            new_theta_value = np.dot(X_oos[t],t_params[-len(self.X_names):]) + t_params[-(len(self.X_names)*2)-1]*np.exp(new_lambda_value/2.0)
            lmda_exp = np.append(lmda_exp,[new_lambda_value]) # For indexing consistency
            scores_exp = np.append(scores_exp,[0]) # expectation of score is zero
            Y_exp = np.append(Y_exp,new_theta_value)

        return lmda_exp

    def _sim_prediction(self, lmda, Y, scores, h, t_params, simulations, X_oos):
        """ Simulates a h-step ahead mean prediction

        Parameters
        ----------
        lmda : np.array
            The past predicted values

        Y : np.array
            The past data

        scores : np.array
            The past scores

        h : int
            How many steps ahead for the prediction

        t_params : np.array
            A vector of (transformed) latent variables

        simulations : int
            How many simulations to perform

        X_oos : np.array
            Out of sample predictors

        Returns
        ----------
        Matrix of simulations
        """     

        sim_vector = np.zeros([simulations,h])

        for n in range(0,simulations):
            # Create arrays to iteratre over
            lmda_exp = lmda.copy()
            scores_exp = scores.copy()
            Y_exp = Y.copy()

            # Loop over h time periods          
            for t in range(0,h):
                new_lambda_value = 0

                if self.p != 0:
                    for j in range(self.p):
                        new_lambda_value += t_params[j]*lmda_exp[-j-1]

                if self.q != 0:
                    for k in range(self.q):
                        new_lambda_value += t_params[k+self.p]*scores_exp[-k-1]

                if self.leverage is True:
                    new_lambda_value += t_params[self.p+self.q]*np.sign(-(Y_exp[-1]-new_theta_value))*(scores_exp[-1]+1)

                new_lambda_value += np.dot(X_oos[t],t_params[-len(self.X_names)*2:-len(self.X_names)]) 
                new_theta_value = np.dot(X_oos[t],t_params[-len(self.X_names):]) + t_params[-(len(self.X_names)*2)-1]*np.exp(new_lambda_value/2.0)

                lmda_exp = np.append(lmda_exp,[new_lambda_value]) # For indexing consistency
                scores_exp = np.append(scores_exp,scores[np.random.randint(scores.shape[0])]) # expectation of score is zero
                Y_exp = np.append(Y_exp,Y[np.random.randint(Y.shape[0])]) # bootstrap returns

            sim_vector[n] = lmda_exp[-h:]

        return np.transpose(sim_vector)

    def _sim_prediction_bayes(self, h, simulations, X_oos):
        """ Simulates a h-step ahead mean prediction

        Parameters
        ----------
        h : int
            How many steps ahead for the prediction

        simulations : int
            How many simulations to perform

        X_oos : np.array
            Out of sample predictors

        Returns
        ----------
        Matrix of simulations
        """     

        sim_vector = np.zeros([simulations,h])

        for n in range(0,simulations):

            t_z = self.draw_latent_variables(nsims=1).T[0]
            lmda, Y, scores, theta = self._model(beta)
            t_z = np.array([self.latent_variables.z_list[k].prior.transform(t_z[k]) for k in range(t_z.shape[0])])

            # Create arrays to iteratre over
            lmda_exp = lmda.copy()
            scores_exp = scores.copy()
            Y_exp = Y.copy()

            # Loop over h time periods          
            for t in range(0,h):
                new_lambda_value = 0

                if self.p != 0:
                    for j in range(self.p):
                        new_lambda_value += t_z[j]*lmda_exp[-j-1]

                if self.q != 0:
                    for k in range(self.q):
                        new_lambda_value += t_z[k+self.p]*scores_exp[-k-1]

                if self.leverage is True:
                    new_lambda_value += t_z[self.p+self.q]*np.sign(-(Y_exp[-1]-new_theta_value))*(scores_exp[-1]+1)

                new_lambda_value += np.dot(X_oos[t],t_z[-len(self.X_names)*2:-len(self.X_names)]) 
                new_theta_value = np.dot(X_oos[t],t_z[-len(self.X_names):]) + t_z[-(len(self.X_names)*2)-1]*np.exp(new_lambda_value/2.0)

                lmda_exp = np.append(lmda_exp,[new_lambda_value]) # For indexing consistency
                scores_exp = np.append(scores_exp,scores[np.random.randint(scores.shape[0])]) # expectation of score is zero
                Y_exp = np.append(Y_exp,Y[np.random.randint(Y.shape[0])]) # bootstrap returns

            sim_vector[n] = lmda_exp[-h:]

        return np.transpose(sim_vector)

    def _sim_predicted_mean(self, lmda, Y, scores, h, t_params, simulations, X_oos):
        """ Simulates a h-step ahead mean prediction

        Parameters
        ----------
        lmda : np.array
            The past predicted values

        Y : np.array
            The past data

        scores : np.array
            The past scores

        h : int
            How many steps ahead for the prediction

        t_params : np.array
            A vector of (transformed) latent variables

        simulations : int
            How many simulations to perform

        X_oos : np.array
            Out of sample predictors

        Returns
        ----------
        Matrix of simulations
        """     

        sim_vector = np.zeros([simulations,h])

        for n in range(0,simulations):
            # Create arrays to iteratre over
            lmda_exp = lmda.copy()
            scores_exp = scores.copy()
            Y_exp = Y.copy()

            # Loop over h time periods          
            for t in range(0,h):
                new_lambda_value = 0

                if self.p != 0:
                    for j in range(self.p):
                        new_lambda_value += t_params[j]*lmda_exp[-j-1]

                if self.q != 0:
                    for k in range(self.q):
                        new_lambda_value += t_params[k+self.p]*scores_exp[-k-1]

                if self.leverage is True:
                    new_lambda_value += t_params[self.p+self.q]*np.sign(-(Y_exp[-1]-new_theta_value))*(scores_exp[-1]+1)

                new_lambda_value += np.dot(X_oos[t],t_params[-len(self.X_names)*2:-len(self.X_names)]) 
                new_theta_value = np.dot(X_oos[t],t_params[-len(self.X_names):]) + t_params[-(len(self.X_names)*2)-1]*np.exp(new_lambda_value/2.0)

                lmda_exp = np.append(lmda_exp,[new_lambda_value]) # For indexing consistency
                scores_exp = np.append(scores_exp,scores[np.random.randint(scores.shape[0])]) # expectation of score is zero
                Y_exp = np.append(Y_exp,Y[np.random.randint(Y.shape[0])]) # bootstrap returns

            sim_vector[n] = lmda_exp[-h:]

        return np.append(lmda, np.array([np.mean(i) for i in np.transpose(sim_vector)]))

    def _summarize_simulations(self, lmda, sim_vector, date_index, h, past_values):
        """ Summarizes a simulation vector and a mean vector of predictions
        
        Parameters
        ----------
        lmda : np.array
            Past volatility values for the moedl
        
        sim_vector : np.array
            N simulation predictions for h-step ahead forecasts
        
        date_index : pd.DateIndex or np.array
            Dates for the simulations
        
        h : int
            How many steps ahead are forecast
        
        past_values : int
            How many past observations to include in the forecast plot
        
        intervals : Boolean
            Would you like to show prediction intervals for the forecast?
        """ 
        mean_values = np.append(lmda, np.array([np.mean(i) for i in sim_vector]))

        error_bars = []
        for pre in range(5,100,5):
            error_bars.append(np.insert([np.percentile(i,pre) for i in sim_vector], 0, mean_values[-h-1]))
        forecasted_values = np.insert([np.mean(i) for i in sim_vector], 0, mean_values[-h-1])
        plot_values = mean_values[-h-past_values:]
        plot_index = date_index[-h-past_values:]

        return error_bars, forecasted_values, plot_values, plot_index

    def add_leverage(self):
        """ Adds leverage term to the model

        Returns
        ----------
        None (changes instance attributes)
        """             

        if self.leverage is True:
            pass
        else:
            self.leverage = True
            self.z_no += 1
            
            for i in range(len(self.X_names)*2+3):
                self.latent_variables.z_list.pop()            

            for parm in range(len(self.X_names)):
                self.latent_variables.add_z('Vol Beta ' + self.X_names[parm], fam.Normal(0,10,transform=None), fam.Normal(0,3))

            for parm in range(len(self.X_names)):
                self.latent_variables.add_z('Returns Beta ' + self.X_names[parm], fam.Normal(0,10,transform=None), fam.Normal(0,3))

            self.latent_variables.add_z('Leverage Term', fam.Flat(transform=None), fam.Normal(0,3))
            self.latent_variables.add_z('v', fam.Flat(transform='exp'), fam.Normal(0,3))
            self.latent_variables.add_z('Returns Constant', fam.Normal(0,3,transform=None), fam.Normal(0,3))
            self.latent_variables.add_z('GARCH-M', fam.Normal(0,3,transform=None), fam.Normal(0,3))
            self.latent_variables.z_list[-3].start = 2.0

    def neg_loglik(self, beta):
        """ Creates the negative log-likelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        The negative logliklihood of the model
        """     

        lmda, Y, _, theta = self._model(beta)
        return -np.sum(ss.t.logpdf(x=Y,
            df=self.latent_variables.z_list[-(len(self.X_names)*2)-2].prior.transform(beta[-(len(self.X_names)*2)-2]),
            loc=theta,scale=np.exp(lmda/2.0)))
    
    def mb_neg_loglik(self, beta, mini_batch):
        """ Creates the negative log-likelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        mini_batch : int
            Size of each mini batch of data

        Returns
        ----------
        The negative logliklihood of the model
        """     
        lmda, Y, _, theta = self._mb_model(beta, mini_batch)
        return -np.sum(ss.t.logpdf(x=Y,
            df=self.latent_variables.z_list[-(len(self.X_names)*2)-2].prior.transform(beta[-(len(self.X_names)*2)-2]),
            loc=theta,scale=np.exp(lmda/2.0)))
    
    def plot_fit(self, **kwargs):
        """ Plots the fit of the model

        Returns
        ----------
        None (plots data and the fit)
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            t_params = self.transform_z()
            plt.figure(figsize=figsize)
            date_index = self.index[max(self.p,self.q):]
            sigma2, Y, ___, theta = self._model(self.latent_variables.get_z_values())
            plt.plot(date_index,np.abs(Y-theta),label=self.data_name + ' Absolute Demeaned Values')
            plt.plot(date_index,np.exp(sigma2/2.0),label='EGARCHMREG(' + str(self.p) + ',' + str(self.q) + ') Conditional Volatility',c='black')                   
            plt.title(self.data_name + " Volatility Plot")  
            plt.legend(loc=2)   
            plt.show()              

    def plot_predict(self, h=5, past_values=20, intervals=True, oos_data=None, **kwargs):

        """ Plots forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show prediction intervals for the forecast?

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        Returns
        ----------
        - Plot of the forecast
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            # Retrieve data, dates and (transformed) latent variables
            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            X_pred = X_oos[:h]
            lmda, Y, scores = self._model(self.latent_variables.get_z_values())           
            date_index = self.shift_dates(h)

            if self.latent_variables.estimation_method in ['M-H']:
                sim_vector = self._sim_prediction_bayes(h, 15000, X_pred)
                error_bars = []

                for pre in range(5,100,5):
                    error_bars.append(np.insert([np.percentile(i,pre) for i in sim_vector], 0, lmda[-1]))

                forecasted_values = np.insert([np.mean(i) for i in sim_vector], 0, lmda[-1])
                plot_values = np.append(lmda[-1-past_values:-2], forecasted_values)
                plot_index = date_index[-h-past_values:]

            else:
                t_z = self.transform_z()
                sim_values = self._sim_prediction(lmda, Y, scores, h, t_z, 15000, X_pred)
                error_bars, forecasted_values, plot_values, plot_index = self._summarize_simulations(lmda, sim_values, date_index, h, past_values)

            plt.figure(figsize=figsize)
            if intervals == True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                for count, pre in enumerate(error_bars):
                    plt.fill_between(date_index[-h-1:], np.exp(error_bars[count]/2.0), np.exp(error_bars[-count-1]/2.0), alpha=alpha[count])   

            plt.plot(plot_index, np.exp(plot_values/2.0))
            plt.title("Forecast for " + self.data_name + " Conditional Volatility")
            plt.xlabel("Time")
            plt.ylabel(self.data_name + " Conditional Volatility")
            plt.show()

    def predict_is(self, h=5, fit_once=True, fit_method='MLE', intervals=False, **kwargs):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        fit_once : boolean
            (default: True) Fits only once before the in-sample prediction; if False, fits after every new datapoint

        fit_method : string
            Which method to fit the model with

        intervals: boolean
            Whether to return prediction intervals

        Returns
        ----------
        - pd.DataFrame with predicted values
        """  

        predictions = []

        for t in range(0, h):
            data1 = self.data_original.iloc[:-(h+t),:]
            data2 = self.data_original.iloc[-h+t:,:]
            x = EGARCHMReg(p=self.p, q=self.q, data=self.data_original[:(-h+t)],formula=self.formula)

            if fit_once is False:
                x.fit(method=fit_method, printer=False)

            if t == 0:
                if fit_once is True:
                    x.fit(method=fit_method, printer=False)
                    saved_lvs = x.latent_variables
                predictions = x.predict(h=1, intervals=intervals, oos_data=data2)
            else:
                if fit_once is True:
                    x.latent_variables = saved_lvs
                predictions = pd.concat([predictions,x.predict(h=1, intervals=intervals, oos_data=data2)])
        
        predictions.rename(columns={0:self.y_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self, h=5, fit_once=True, fit_method='MLE', **kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        fit_once : boolean
            (default: True) Fits only once before the in-sample prediction; if False, fits after every new datapoint

        fit_method : string
            Which method to fit the model with

        Returns
        ----------
        - Plot of the forecast against data 
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize', (10,7))

        plt.figure(figsize=figsize)
        date_index = self.index[-h:]
        predictions = self.predict_is(h, fit_method=fit_method, fit_once=fit_once)
        data = self.data[-h:]

        t_params = self.transform_z()

        plt.plot(date_index, np.abs(data-t_params[-1]), label='Data')
        plt.plot(date_index, predictions, label='Predictions', c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def predict(self, h=5, oos_data=None, intervals=False):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        oos_data : pd.DataFrame
            Data to use for the predictors in the forecast

        intervals : boolean (default: False)
            Whether to return prediction intervals

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     
        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            X_pred = X_oos[:h]
            lmda, Y, scores = self._model(self.latent_variables.get_z_values())         
            date_index = self.shift_dates(h)

            if self.latent_variables.estimation_method in ['M-H']:
                sim_vector = self._sim_prediction_bayes(h, 15000, X_pred)
                error_bars = []

                for pre in range(5,100,5):
                    error_bars.append(np.insert([np.percentile(i,pre) for i in sim_vector], 0, lmda[-1]))

                forecasted_values = np.array([np.mean(i) for i in sim_vector])
                prediction_01 = np.array([np.percentile(i, 1) for i in sim_vector])
                prediction_05 = np.array([np.percentile(i, 5) for i in sim_vector])
                prediction_95 = np.array([np.percentile(i, 95) for i in sim_vector])
                prediction_99 = np.array([np.percentile(i, 99) for i in sim_vector])

            else:
                t_z = self.transform_z()

                if intervals is True:
                    sim_values = self._sim_prediction(lmda, Y, scores, h, t_z, 15000, X_pred)
                else:
                    sim_values = self._sim_prediction(lmda, Y, scores, h, t_z, 2, X_pred)

                mean_values = self._sim_predicted_mean(lmda, Y, scores, h, t_z, 15000, X_pred)
                forecasted_values = mean_values[-h:]

            if intervals is False:
                result = pd.DataFrame(np.exp(forecasted_values/2.0))
                result.rename(columns={0:self.data_name}, inplace=True)
            else:
                if self.latent_variables.estimation_method not in ['M-H']:
                    sim_values = self._sim_prediction(lmda, Y, scores, h, t_z, 15000, X_pred)
                    prediction_01 = np.array([np.percentile(i, 1) for i in sim_values])
                    prediction_05 = np.array([np.percentile(i, 5) for i in sim_values])
                    prediction_95 = np.array([np.percentile(i, 95) for i in sim_values])
                    prediction_99 = np.array([np.percentile(i, 99) for i in sim_values])

                result = np.exp(pd.DataFrame([forecasted_values, prediction_01, prediction_05, 
                    prediction_95, prediction_99]).T/2.0)
                result.rename(columns={0:self.data_name, 1: "1% Prediction Interval", 
                    2: "5% Prediction Interval", 3: "95% Prediction Interval", 4: "99% Prediction Interval"}, 
                    inplace=True)
 
            result.index = date_index[-h:]

            return result

    def sample(self, nsims=1000):
        """ Samples from the posterior predictive distribution

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws from the posterior predictive distribution

        Returns
        ----------
        - np.ndarray of draws from the data
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            lv_draws = self.draw_latent_variables(nsims=nsims)
            sigmas = [self._model(lv_draws[:,i])[0] for i in range(nsims)]
            data_draws = np.array([ss.t.rvs(loc=self.latent_variables.z_list[-len(self.X_names)*2-1].prior.transform(lv_draws[-len(self.X_names)*2-1,i]),
                df=self.latent_variables.z_list[-len(self.X_names)*2-2].prior.transform(lv_draws[-len(self.X_names)*2-2,i]), scale=np.exp(sigmas[i]/2.0)) for i in range(nsims)])
            return data_draws

    def plot_sample(self, nsims=10, plot_data=True, **kwargs):
        """
        Plots draws from the posterior predictive density against the data

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws from the posterior predictive distribution

        plot_data boolean
            Whether to plot the data or not
        """

        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            import matplotlib.pyplot as plt
            import seaborn as sns

            alpha = 1.0
            figsize = kwargs.get('figsize',(10,7))
            plt.figure(figsize=figsize)
            date_index = self.index[max(self.p,self.q):]
            sigma2, Y, scores, _ = self._model(self.latent_variables.get_z_values()) 
            draws = self.sample(nsims).T
            plt.plot(date_index, draws, label='Posterior Draws', alpha=1.0)
            if plot_data is True:
                plt.plot(date_index, Y, label='Data', c='black', alpha=0.5, linestyle='', marker='s')
            plt.title(self.data_name)
            plt.show()    

    def ppc(self, nsims=1000, T=np.mean):
        """ Computes posterior predictive p-value

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws for the PPC

        T : function
            A discrepancy measure - e.g. np.mean, np.std, np.max

        Returns
        ----------
        - float (posterior predictive p-value)
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            lv_draws = self.draw_latent_variables(nsims=nsims)
            sigmas = [self._model(lv_draws[:,i])[0] for i in range(nsims)]
            data_draws = np.array([ss.t.rvs(loc=self.latent_variables.z_list[-len(self.X_names)*2-1].prior.transform(lv_draws[-len(self.X_names)*2-1,i]),
                df=self.latent_variables.z_list[-len(self.X_names)*2-2].prior.transform(lv_draws[-len(self.X_names)*2-2,i]), scale=np.exp(sigmas[i]/2.0)) for i in range(nsims)])
            T_sims = T(self.sample(nsims=nsims), axis=1)
            T_actual = T(self.data)
            return len(T_sims[T_sims>T_actual])/nsims

    def plot_ppc(self, nsims=1000, T=np.mean, **kwargs):
        """ Plots histogram of the discrepancy from draws of the posterior

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws for the PPC

        T : function
            A discrepancy measure - e.g. np.mean, np.std, np.max
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            import matplotlib.pyplot as plt
            import seaborn as sns

            figsize = kwargs.get('figsize',(10,7))

            lv_draws = self.draw_latent_variables(nsims=nsims)
            sigmas = [self._model(lv_draws[:,i])[0] for i in range(nsims)]
            data_draws = np.array([ss.t.rvs(loc=self.latent_variables.z_list[-len(self.X_names)*2-1].prior.transform(lv_draws[-len(self.X_names)*2-1,i]),
                df=self.latent_variables.z_list[-len(self.X_names)*2-2].prior.transform(lv_draws[-len(self.X_names)*2-2,i]), scale=np.exp(sigmas[i]/2.0)) for i in range(nsims)])
            T_sim = T(self.sample(nsims=nsims), axis=1)
            T_actual = T(self.data)

            if T == np.mean:
                description = " of the mean"
            elif T == np.max:
                description = " of the maximum"
            elif T == np.min:
                description = " of the minimum"
            elif T == np.median:
                description = " of the median"
            else:
                description = ""

            plt.figure(figsize=figsize)
            ax = plt.subplot()
            ax.axvline(T_actual)
            sns.distplot(T_sim, kde=False, ax=ax)
            ax.set(title='Posterior predictive' + description, xlabel='T(x)', ylabel='Frequency');
            plt.show()


import numpy as np
import scipy.stats as ss
import scipy.special as sp

from .family import Family


class InverseGamma(Family):
    """ 
    Inverse Gamma Distribution
    ----
    This class contains methods relating to the inverse gamma distribution for time series.
    """

    def __init__(self, alpha, beta, transform=None, **kwargs):
        """
        Parameters
        ----------
        alpha : float
            Alpha parameter for the Inverse Gamma distribution

        beta : float
            Beta parameter for the Inverse Gamma distribution

        transform : str
            Whether to apply a transformation - e.g. 'exp' or 'logit'
        """
        super(InverseGamma, self).__init__(transform)
        self.covariance_prior = False
        self.alpha = alpha
        self.beta = beta

    def logpdf(self, x):
        """
        Log PDF for Inverse Gamma prior

        Parameters
        ----------
        x : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - log(p(x))
        """
        if self.transform is not None:
            x = self.transform(x)       
        return (-self.alpha-1)*np.log(x) - (self.beta/float(x))

    def pdf(self, x):
        """
        PDF for Inverse Gamma prior

        Parameters
        ----------
        x : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - p(x)
        """
        if self.transform is not None:
            x = self.transform(x)               
        return (x**(-self.alpha-1))*np.exp(-(self.beta/float(x)))


import numpy as np

from .output import TablePrinter
from .tests import find_p_value
from .inference import norm_post_sim

class Results(object):

    def __init__(self):
        pass


class MLEResults(Results):

    def __init__(self,data_name,X_names,model_name,model_type,latent_variables, 
        results,data,index, multivariate_model,objective_object,method,
        z_hide,max_lag,ihessian=None,signal=None,scores=None,states=None,
        states_var=None):

        self.data_name = data_name
        self.X_names = X_names
        self.max_lag = max_lag
        self.model_name = model_name
        self.model_type = model_type
        self.z = latent_variables
        self.z_values = latent_variables.get_z_values()
        self.results = results
        self.method = method

        self.ihessian = ihessian
        self.scores = scores
        self.states = states
        self.states_var = states_var
        self.data = data
        self.index = index
        self.signal = signal
        self.multivariate_model = multivariate_model
        self.z_hide = int(z_hide)

        if self.multivariate_model is True:
            self.data_length = self.data[0].shape[0]
            self.data_name = ",".join(self.data_name)
        else:
            self.data_length = self.data.shape[0]

        self.objective_object = objective_object

        if self.method == 'MLE' or self.method == 'OLS':
            self.loglik = -self.objective_object(self.z_values)
            self.aic = 2*len(self.z_values)+2*self.objective_object(self.z_values)
            self.bic = 2*self.objective_object(self.z_values) + len(self.z_values)*np.log(self.data_length)
        elif self.method == 'PML':
            self.aic = 2*len(self.z_values)+2*self.objective_object(self.z_values)
            self.bic = 2*self.objective_object(self.z_values) + len(self.z_values)*np.log(self.data_length)

        if self.model_type in ['LLT','LLEV']:
            self.rounding_points = 10
        else:
            self.rounding_points = 4

    def __str__(self):
        if self.method == 'MLE':
            print("MLE Results Object")
        elif self.method == 'OLS':
            print("OLS Results Object")
        else:
            print("PML Results Object")
        print("==========================")
        print("Dependent variable: " + self.data_name)
        print("Regressors: " + str(self.X_names))
        print("==========================")
        print("Latent Variable Attributes: ")
        if self.ihessian is not None:
            print(".ihessian: Inverse Hessian")     
        print(".z : LatentVariables() object")
        if self.results is not None:
            print(".results : optimizer results")
        print("")
        print("Implied Model Attributes: ")
        print(".aic: Akaike Information Criterion") 
        print(".bic: Bayesian Information Criterion")   
        print(".data: Model Data")  
        print(".index: Model Index")
        if self.method == 'MLE' or self.method == 'OLS':
            print(".loglik: Loglikelihood") 
        if self.scores is not None:
            print(".scores: Model Scores")                      
        if self.signal is not None:
            print(".signal: Model Signal")          
        if self.states is not None:
            print(".states: Model States")      
        if self.states_var is not None:
            print(".states_var: Model State Variances")                                     
        print(".results : optimizer results")
        print("")       
        print("Methods: ")
        print(".summary() : printed results")
        return("")

    def summary(self, transformed=True):
        if self.ihessian is not None:
            return self.summary_with_hessian(transformed)
        else:
            return self.summary_without_hessian()

    def summary_with_hessian(self, transformed=True):

        ses = np.power(np.abs(np.diag(self.ihessian)),0.5)
        t_z = self.z.get_z_values(transformed=True)
        t_p_std = ses.copy() # vector for transformed standard errors

        # Create transformed variables
        # for k in range(len(t_z)-self.z_hide):
        #    z_temp = (self.z_values[k]/float(ses[k]))
        #    t_p_std[k] = t_z[k] / z_temp

        data = []

        for i in range(len(self.z.z_list)-int(self.z_hide)):
            if self.z.z_list[i].prior.transform == np.array:
                data.append({
                    'z_name': self.z.z_list[i].name, 
                    'z_value':np.round(self.z.z_list[i].prior.transform(self.z_values[i]),self.rounding_points), 
                    'z_std': np.round(t_p_std[i],self.rounding_points),
                    'z_z': np.round(t_z[i]/float(t_p_std[i]),self.rounding_points),
                    'z_p': np.round(find_p_value(t_z[i]/float(t_p_std[i])),self.rounding_points),
                    'ci': "(" + str(np.round(t_z[i] - t_p_std[i]*1.96,self.rounding_points)) + " | " + str(np.round(t_z[i] + t_p_std[i]*1.96,self.rounding_points)) + ")"})
            else:
                if transformed is True:
                    data.append({
                        'z_name': self.z.z_list[i].name, 
                        'z_value':np.round(self.z.z_list[i].prior.transform(self.z_values[i]),self.rounding_points)})        
                else:
                    data.append({
                        'z_name': self.z.z_list[i].prior.itransform_name + '(' + self.z.z_list[i].name + ')', 
                        'z_value':np.round(self.z_values[i],self.rounding_points), 
                        'z_std': np.round(t_p_std[i],self.rounding_points),
                        'z_z': np.round(t_z[i]/float(t_p_std[i]),self.rounding_points),
                        'z_p': np.round(find_p_value(t_z[i]/float(t_p_std[i])),self.rounding_points),
                        'ci': "(" + str(np.round(t_z[i] - t_p_std[i]*1.96,self.rounding_points)) + " | " + str(np.round(t_z[i] + t_p_std[i]*1.96,self.rounding_points)) + ")"})       
        
        fmt = [
            ('Latent Variable',       'z_name',   40),
            ('Estimate',          'z_value',       10),
            ('Std Error', 'z_std', 10),
            ('z',          'z_z',       8),
            ('P>|z|',          'z_p',       8),
            ('95% C.I.',          'ci',       25)
            ]

        model_details = []

        model_fmt = [
            (self.model_name, 'model_details', 55),
            ('', 'model_results', 50)
            ]

        if self.method == 'MLE' or self.method == 'OLS' :
            obj_desc = "Log Likelihood: " + str(np.round(-self.objective_object(self.z_values),4))
        else:
            obj_desc = "Unnormalized Log Posterior: " + str(np.round(-self.objective_object(self.z_values),4))

        model_details.append({'model_details': 'Dependent Variable: ' + str(self.data_name), 
            'model_results': 'Method: ' + str(self.method)})
        model_details.append({'model_details': 'Start Date: ' + str(self.index[self.max_lag]),
            'model_results': obj_desc})
        model_details.append({'model_details': 'End Date: ' + str(self.index[-1]),
            'model_results': 'AIC: ' + str(np.round(2*len(self.z_values)+2*self.objective_object(self.z_values),4))})
        model_details.append({'model_details': 'Number of observations: ' + str(self.data_length),
            'model_results': 'BIC: ' + str(np.round(2*self.objective_object(self.z_values) + len(self.z_values)*np.log(self.data_length),4))})


        print( TablePrinter(model_fmt, ul='=')(model_details) )
        print("="*106)
        print( TablePrinter(fmt, ul='=')(data) )
        print("="*106)
        if 'Skewt' in self.model_name:
            print("WARNING: Skew t distribution is not well-suited for MLE or MAP inference")
            print("Workaround 1: Use a t-distribution instead for MLE/MAP")
            print("Workaround 2: Use M-H or BBVI inference for Skew t distribution")

    def summary_without_hessian(self):

        t_z = self.z.get_z_values(transformed=True)

        print ("Hessian not invertible! Consider a different model specification.")
        print ("")      

        data = []

        for i in range(len(self.z.z_list)):
            data.append({'z_name': self.z.z_list[i].name, 'z_value':np.round(self.z.z_list[i].prior.transform(self.results.x[i]),4)})

        fmt = [
            ('Latent Variable',       'z_name',   40),
            ('Estimate',          'z_value',       10)
            ]

        model_details = []

        model_fmt = [
            (self.model_name, 'model_details', 55),
            ('', 'model_results', 50)
            ]

        if self.method == 'MLE':
            obj_desc = "Log Likelihood: " + str(np.round(-self.objective_object(self.results.x),4))
        else:
            obj_desc = "Unnormalized Log Posterior: " + str(np.round(-self.objective_object(self.results.x),4))

        model_details.append({'model_details': 'Dependent Variable: ' + self.data_name, 
            'model_results': 'Method: ' + str(self.method)})
        model_details.append({'model_details': 'Start Date: ' + str(self.index[self.max_lag]),
            'model_results': obj_desc})
        model_details.append({'model_details': 'End Date: ' + str(self.index[-1]),
            'model_results': 'AIC: ' + str(self.aic)})
        model_details.append({'model_details': 'Number of observations: ' + str(self.data_length),
            'model_results': 'BIC: ' + str(self.bic)})


        print( TablePrinter(model_fmt, ul='=')(model_details) )
        print("="*106)
        print( TablePrinter(fmt, ul='=')(data) )
        print("="*106)
        if 'Skewt' in self.model_name:
            print("WARNING: Skew t distribution is not well-suited for MLE or MAP inference")
            print("Workaround 1: Use a t-distribution instead for MLE/MAP")
            print("Workaround 2: Use M-H or BBVI inference for Skew t distribution")


class BBVIResults(Results):

    def __init__(self, data_name, X_names, model_name, model_type, latent_variables, 
        data,index, multivariate_model, objective_object, method, z_hide, max_lag, ses,
        signal=None, scores=None, elbo_records=None, states=None, states_var=None):

        self.data_name = data_name
        self.X_names = X_names
        self.max_lag = max_lag
        self.model_name = model_name
        self.model_type = model_type
        self.z = latent_variables
        self.method = method

        self.ses = ses
        self.scores = scores
        self.states = states
        self.states_var = states_var
        self.data = data
        self.index = index
        self.signal = signal
        self.multivariate_model = multivariate_model
        self.z_hide = z_hide
        self.elbo_records = elbo_records

        if self.multivariate_model is True:
            self.data_length = self.data[0].shape[0]
            self.data_name = ",".join(self.data_name)
        else:
            self.data_length = self.data.shape[0]

        z_values = self.z.get_z_values(transformed=False)

        self.objective_object = objective_object
        self.aic = 2*len(z_values)+2*self.objective_object(z_values)
        self.bic = 2*self.objective_object(z_values) + len(z_values)*np.log(self.data_length)

        if self.model_type in ['LLT','LLEV']:
            self.rounding_points = 10
        else:
            self.rounding_points = 4

    def __str__(self):
        print("BBVI Results Object")
        print("==========================")
        print("Dependent variable: " + self.data_name)
        print("Regressors: " + str(self.X_names))
        print("==========================")
        print("Latent Variables Attributes: ")
        print(".z : LatentVariables() object")
        print(".results : optimizer results")
        print("")
        print("Implied Model Attributes: ")
        print(".aic: Akaike Information Criterion") 
        print(".bic: Bayesian Information Criterion")   
        print(".data: Model Data")  
        print(".index: Model Index")
        if self.scores is not None:
            print(".scores: Model Scores")                      
        if self.signal is not None:
            print(".signal: Model Signal")          
        if self.states is not None:
            print(".states: Model States")  
        if self.states_var is not None:
            print(".states_var: Model State Variances")                                     
        print("")       
        print("Methods: ")
        print(".summary() : printed results")
        return("")

    def plot_elbo(self, figsize=(15,7)):
        """
        Plots the ELBO progress (if present)
        """
        import matplotlib.pyplot as plt

        plt.figure(figsize=figsize)
        plt.plot(self.elbo_records)
        plt.xlabel("Iterations")
        plt.ylabel("ELBO")
        plt.show()

    def summary(self, transformed=True):
        ihessian = np.diag(np.power(np.exp(self.ses),2))
        z_values = self.z.get_z_values(transformed=False)
        chain, mean_est, median_est, upper_95_est, lower_95_est = norm_post_sim(z_values,ihessian)

        if transformed is True:
            for k in range(len(chain)):
                chain[k] = self.z.z_list[k].prior.transform(chain[k])
                mean_est[k] = self.z.z_list[k].prior.transform(mean_est[k])
                median_est[k] = self.z.z_list[k].prior.transform(median_est[k])
                upper_95_est[k] = self.z.z_list[k].prior.transform(upper_95_est[k])
                lower_95_est[k] = self.z.z_list[k].prior.transform(lower_95_est[k])                

        mean_est = np.array(mean_est)
        self.chains = chain[:]

        data = []

        for i in range(len(self.z.z_list)-int(self.z_hide)):
            data.append({
                'z_name': self.z.z_list[i].name, 
                'z_mean': np.round(mean_est[i],self.rounding_points),
                'z_median': np.round(median_est[i],self.rounding_points),
                'ci':  "(" + str(np.round(lower_95_est[i],self.rounding_points)) + " | " + str(np.round(upper_95_est[i],self.rounding_points)) + ")"
                })                  

        fmt = [
            ('Latent Variable','z_name',40),
            ('Median','z_median',18),
            ('Mean', 'z_mean', 18),
            ('95% Credibility Interval','ci',25)]

        model_details = []

        model_fmt = [
            (self.model_name, 'model_details', 55),
            ('', 'model_results', 50)
            ]

        if self.method == 'MLE':
            obj_desc = "Log Likelihood: " + str(np.round(-self.objective_object(z_values),4))
        else:
            obj_desc = "Unnormalized Log Posterior: " + str(np.round(-self.objective_object(z_values),4))

        model_details.append({'model_details': 'Dependent Variable: ' + self.data_name, 
            'model_results': 'Method: ' + str(self.method)})
        model_details.append({'model_details': 'Start Date: ' + str(self.index[self.max_lag]),
            'model_results': obj_desc})
        model_details.append({'model_details': 'End Date: ' + str(self.index[-1]),
            'model_results': 'AIC: ' + str(self.aic)})
        model_details.append({'model_details': 'Number of observations: ' + str(self.data_length),
            'model_results': 'BIC: ' + str(self.bic)})

        print( TablePrinter(model_fmt, ul='=')(model_details) )
        print("="*106)
        print( TablePrinter(fmt, ul='=')(data) )
        print("="*106)


class BBVISSResults(Results):

    def __init__(self,data_name,X_names,model_name,model_type,latent_variables, 
        data,index,multivariate_model,objective,method,
        z_hide,max_lag,ses,signal=None,scores=None,states=None,
        states_var=None,elbo_records=None):

        self.data_name = data_name
        self.X_names = X_names
        self.max_lag = max_lag
        self.model_name = model_name
        self.model_type = model_type
        self.z = latent_variables
        self.method = method

        self.ses = ses
        self.scores = scores
        self.states = states
        self.states_var = states_var
        self.data = data
        self.index = index
        self.signal = signal
        self.multivariate_model = multivariate_model
        self.z_hide = z_hide
        self.elbo_records = elbo_records

        if self.multivariate_model is True:
            self.data_length = self.data[0].shape[0]
            self.data_name = ",".join(self.data_name)
        else:
            self.data_length = self.data.shape[0]

        z_values = self.z.get_z_values(transformed=False)

        self.objective = objective
        self.aic = 2*len(z_values)+2*self.objective
        self.bic = 2*self.objective + len(z_values)*np.log(self.data_length)

        if self.model_type in ['LLT','LLEV']:
            self.rounding_points = 10
        else:
            self.rounding_points = 4

    def __str__(self):
        print("BBVI Results Object")
        print("==========================")
        print("Dependent variable: " + self.data_name)
        print("Regressors: " + str(self.X_names))
        print("==========================")
        print("Latent Variable Attributes: ")
        print(".z : LatentVariables() object")
        print("")
        print("Implied Model Attributes: ")
        print(".aic: Akaike Information Criterion") 
        print(".bic: Bayesian Information Criterion")   
        print(".data: Model Data")  
        print(".index: Model Index")
        print(".objective: Unnormalized Log Posterior")
        if self.scores is not None:
            print(".scores: Model Scores")                      
        if self.signal is not None:
            print(".signal: Model Signal")          
        if self.states is not None:
            print(".states: Model States")  
        if self.states_var is not None:
            print(".states_var: Model State Variances")                                     
        print("")       
        print("Methods: ")
        print(".summary() : printed results")
        return("")

    def plot_elbo(self, figsize=(15,7)):
        """
        Plots the ELBO progress (if present)
        """
        import matplotlib.pyplot as plt

        plt.figure(figsize=figsize)
        plt.plot(self.elbo_records)
        plt.xlabel("Iterations")
        plt.ylabel("ELBO")
        plt.show()

    def summary(self, transformed=True):
        ihessian = np.diag(np.power(np.exp(self.ses),2))
        z_values = self.z.get_z_values(transformed=False)
        chain, mean_est, median_est, upper_95_est, lower_95_est = norm_post_sim(z_values,ihessian)

        if transformed is True:
            for k in range(len(z_values)):
                chain[k] = self.z.z_list[k].prior.transform(chain[k])
                mean_est[k] = self.z.z_list[k].prior.transform(mean_est[k])
                median_est[k] = self.z.z_list[k].prior.transform(median_est[k])
                upper_95_est[k] = self.z.z_list[k].prior.transform(upper_95_est[k])
                lower_95_est[k] = self.z.z_list[k].prior.transform(lower_95_est[k])                

        mean_est = np.array(mean_est)
        self.chains = chain[:]

        data = []

        for i in range(len(self.z.z_list)-int(self.z_hide)):
            data.append({
                'z_name': self.z.z_list[i].name, 
                'z_mean': np.round(mean_est[i],self.rounding_points),
                'z_median': np.round(median_est[i],self.rounding_points),
                'ci':  "(" + str(np.round(lower_95_est[i],self.rounding_points)) + " | " + str(np.round(upper_95_est[i],self.rounding_points)) + ")"
                })                  

        fmt = [
            ('Latent Variable','z_name',40),
            ('Median','z_median',18),
            ('Mean', 'z_mean', 18),
            ('95% Credibility Interval','ci',25)]

        model_details = []

        model_fmt = [
            (self.model_name, 'model_details', 55),
            ('', 'model_results', 50)
            ]

        obj_desc = "Unnormalized Log Posterior: " + str(np.round(-self.objective,4))

        model_details.append({'model_details': 'Dependent Variable: ' + self.data_name, 
            'model_results': 'Method: ' + str(self.method)})
        model_details.append({'model_details': 'Start Date: ' + str(self.index[self.max_lag]),
            'model_results': obj_desc})
        model_details.append({'model_details': 'End Date: ' + str(self.index[-1]),
            'model_results': 'AIC: ' + str(self.aic)})
        model_details.append({'model_details': 'Number of observations: ' + str(self.data_length),
            'model_results': 'BIC: ' + str(self.bic)})

        print( TablePrinter(model_fmt, ul='=')(model_details) )
        print("="*106)
        print( TablePrinter(fmt, ul='=')(data) )
        print("="*106)


class LaplaceResults(Results):

    def __init__(self,data_name,X_names,model_name,model_type,latent_variables, 
        data,index,multivariate_model,objective_object,method,
        z_hide,max_lag,ihessian,signal=None,scores=None,states=None,
        states_var=None):

        self.data_name = data_name
        self.X_names = X_names
        self.max_lag = max_lag
        self.model_name = model_name
        self.model_type = model_type
        self.z = latent_variables
        self.method = method

        self.ihessian = ihessian
        self.scores = scores
        self.states = states
        self.states_var = states_var
        self.data = data
        self.index = index
        self.signal = signal
        self.multivariate_model = multivariate_model
        self.z_hide = z_hide

        if self.multivariate_model is True:
            self.data_length = self.data[0].shape[0]
            self.data_name = ",".join(self.data_name)
        else:
            self.data_length = self.data.shape[0]

        z_values = self.z.get_z_values(transformed=False)

        self.objective_object = objective_object
        self.aic = 2*len(z_values)+2*self.objective_object(z_values)
        self.bic = 2*self.objective_object(z_values) + len(z_values)*np.log(self.data_length)

        if self.model_type in ['LLT','LLEV']:
            self.rounding_points = 10
        else:
            self.rounding_points = 4

    def __str__(self):
        print("Laplace Results Object")
        print("==========================")
        print("Dependent variable: " + self.data_name)
        print("Regressors: " + str(self.X_names))
        print("==========================")
        print("Latent Variables Attributes: ")
        if self.ihessian is not None:
            print(".ihessian: Inverse Hessian")             
        print(".z : LatentVariables() object")
        print(".results : optimizer results")
        print("")
        print("Implied Model Attributes: ")
        print(".aic: Akaike Information Criterion") 
        print(".bic: Bayesian Information Criterion")   
        print(".data: Model Data")  
        print(".index: Model Index")
        if self.scores is not None:
            print(".scores: Model Scores")                      
        if self.signal is not None:
            print(".signal: Model Signal")          
        if self.states is not None:
            print(".states: Model States")  
        if self.states_var is not None:
            print(".states_var: Model State Variances")                                     
        print("")       
        print("Methods: ")
        print(".summary() : printed results")
        return("")

    def summary(self, transformed=True):
        z_values = self.z.get_z_values(transformed=False)
        chain, mean_est, median_est, upper_95_est, lower_95_est = norm_post_sim(z_values,self.ihessian)

        if transformed is True:
            for k in range(len(chain)):
                chain[k] = self.z.z_list[k].prior.transform(chain[k])
                mean_est[k] = self.z.z_list[k].prior.transform(mean_est[k])
                median_est[k] = self.z.z_list[k].prior.transform(median_est[k])
                upper_95_est[k] = self.z.z_list[k].prior.transform(upper_95_est[k])
                lower_95_est[k] = self.z.z_list[k].prior.transform(lower_95_est[k])                

        mean_est = np.array(mean_est)
        self.chains = chain[:]

        data = []

        for i in range(len(self.z.z_list)-int(self.z_hide)):
            data.append({
                'z_name': self.z.z_list[i].name, 
                'z_mean': np.round(mean_est[i],self.rounding_points),
                'z_median': np.round(median_est[i],self.rounding_points),
                'ci':  "(" + str(np.round(lower_95_est[i],self.rounding_points)) + " | " + str(np.round(upper_95_est[i],self.rounding_points)) + ")"
                })                  

        fmt = [
            ('Latent Variable','z_name',40),
            ('Median','z_median',18),
            ('Mean', 'z_mean', 18),
            ('95% Credibility Interval','ci',25)]

        model_details = []

        model_fmt = [
            (self.model_name, 'model_details', 55),
            ('', 'model_results', 50)
            ]

        if self.method == 'MLE':
            obj_desc = "Log Likelihood: " + str(np.round(-self.objective_object(z_values),4))
        else:
            obj_desc = "Unnormalized Log Posterior: " + str(np.round(-self.objective_object(z_values),4))

        model_details.append({'model_details': 'Dependent Variable: ' + self.data_name, 
            'model_results': 'Method: ' + str(self.method)})
        model_details.append({'model_details': 'Start Date: ' + str(self.index[self.max_lag]),
            'model_results': obj_desc})
        model_details.append({'model_details': 'End Date: ' + str(self.index[-1]),
            'model_results': 'AIC: ' + str(self.aic)})
        model_details.append({'model_details': 'Number of observations: ' + str(self.data_length),
            'model_results': 'BIC: ' + str(self.bic)})

        print( TablePrinter(model_fmt, ul='=')(model_details) )
        print("="*106)
        print( TablePrinter(fmt, ul='=')(data) )
        print("="*106)

class MCMCResults(Results):

    def __init__(self,data_name,X_names,model_name,model_type,latent_variables, 
        data,index,multivariate_model,objective_object,method,
        z_hide,max_lag,samples,mean_est,median_est,lower_95_est,upper_95_est,
        signal=None,scores=None,states=None,states_var=None):
        self.data_name = data_name
        self.X_names = X_names
        self.max_lag = max_lag
        self.model_name = model_name
        self.model_type = model_type
        self.z = latent_variables
        self.method = method

        self.samples = samples
        self.mean_est = mean_est
        self.median_est = median_est
        self.lower_95_est = lower_95_est
        self.upper_95_est = upper_95_est
        self.scores = scores
        self.states = states
        self.states_var = states_var
        self.data = data
        self.index = index
        self.signal = signal
        self.multivariate_model = multivariate_model
        self.z_hide = z_hide

        if self.multivariate_model is True:
            self.data_length = self.data[0].shape[0]
            self.data_name = ",".join(self.data_name)
        else:
            self.data_length = self.data.shape[0]

        z_values = self.z.get_z_values(transformed=False)

        self.objective_object = objective_object
        self.aic = 2*len(z_values)+2*self.objective_object(z_values)
        self.bic = 2*self.objective_object(z_values) + len(z_values)*np.log(self.data_length)

        if self.model_type in ['LLT','LLEV']:
            self.rounding_points = 10
        else:
            self.rounding_points = 4

    def __str__(self):
        print("Metropolis Hastings Results Object")
        print("==========================")
        print("Dependent variable: " + self.data_name)
        print("Regressors: " + str(self.X_names))
        print("==========================")
        print("Latent Variable Attributes: ")     
        print(".z : LatentVariables() object")
        if self.samples is not None:
            print(".samples: MCMC samples")             
        print("")
        print("Implied Model Attributes: ")
        print(".aic: Akaike Information Criterion") 
        print(".bic: Bayesian Information Criterion")   
        print(".data: Model Data")  
        print(".index: Model Index")
        if self.scores is not None:
            print(".scores: Model Scores")                      
        if self.signal is not None:
            print(".signal: Model Signal")          
        if self.states is not None:
            print(".states: Model States")  
        if self.states_var is not None:
            print(".states_var: Model State Variances")                                     
        print("")       
        print("Methods: ")
        print(".summary() : printed results")
        return("")

    def summary(self, transformed=True):
        z_values = self.z.get_z_values(transformed=False)

        data = []

        for i in range(len(self.z.z_list)-int(self.z_hide)):
            data.append({
                'z_name': self.z.z_list[i].name, 
                'z_mean': np.round(self.mean_est[i],self.rounding_points),
                'z_median': np.round(self.median_est[i],self.rounding_points),
                'ci':  "(" + str(np.round(self.lower_95_est[i],self.rounding_points)) + " | " + str(np.round(self.upper_95_est[i],self.rounding_points)) + ")"
                })                  

        fmt = [
            ('Latent Variable','z_name',40),
            ('Median','z_median',18),
            ('Mean', 'z_mean', 18),
            ('95% Credibility Interval','ci',25)]

        model_details = []

        model_fmt = [
            (self.model_name, 'model_details', 55),
            ('', 'model_results', 50)
            ]

        if self.method == 'MLE':
            obj_desc = "Log Likelihood: " + str(np.round(-self.objective_object(z_values),4))
        else:
            obj_desc = "Unnormalized Log Posterior: " + str(np.round(-self.objective_object(z_values),4))

        model_details.append({'model_details': 'Dependent Variable: ' + self.data_name, 
            'model_results': 'Method: ' + str(self.method)})
        model_details.append({'model_details': 'Start Date: ' + str(self.index[self.max_lag]),
            'model_results': obj_desc})
        model_details.append({'model_details': 'End Date: ' + str(self.index[-1]),
            'model_results': 'AIC: ' + str(self.aic)})
        model_details.append({'model_details': 'Number of observations: ' + str(self.data_length),
            'model_results': 'BIC: ' + str(self.bic)})

        print( TablePrinter(model_fmt, ul='=')(model_details) )
        print("="*106)
        print( TablePrinter(fmt, ul='=')(data) )
        print("="*106)



from .llm import LLEV
from .llt import LLT
from .nllm import NLLEV
from .nllt import NLLT
from .dynlin import DynReg
from .dar import DAR
from .ndynlin import NDynReg
from .kalman import nl_univariate_KFS, nld_univariate_KFS
from .local_level import LocalLevel
from .local_trend import LocalTrend
from .dynamic_glm import DynamicGLM


import io
import os
import re
import sys
import subprocess

PACKAGE_NAME = 'pyflux'
DESCRIPTION = "PyFlux: A time-series analysis library for Python"
LONG_DESCRIPTION = DESCRIPTION
AUTHOR = "Ross Taylor"
AUTHOR_EMAIL = "rj-taylor@live.co.uk"
URL = 'https://github.com/rjt1990/pyflux'
DOWNLOAD_URL = 'https://github.com/rjt1990/pyflux/tarball/0.4.10'
LICENSE = 'BSD'

def version(package, encoding='utf-8'):
    """Obtain the packge version from a python file e.g. pkg/__init__.py
    See <https://packaging.python.org/en/latest/single_source_version.html>.
    """
    path = os.path.join(os.path.dirname(__file__), package, '__init__.py')
    with io.open(path, encoding=encoding) as fp:
        version_info = fp.read()
    version_match = re.search(r"""^__version__ = ['"]([^'"]*)['"]""",
                              version_info, re.M)
    if not version_match:
        raise RuntimeError("Unable to find version string.")
    return version_match.group(1)

def generate_cython(package):
    """Cythonize all sources in the package"""
    cwd = os.path.abspath(os.path.dirname(__file__))
    print("Cythonizing sources")
    p = subprocess.call([sys.executable,
                         os.path.join(cwd, 'tools', 'cythonize.py'),
                         package],
                        cwd=cwd)
    if p != 0:
        raise RuntimeError("Running cythonize failed!")

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration(None, parent_package, top_path)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)

    config.add_subpackage(PACKAGE_NAME)

    return config

def setup_package():
    from numpy.distutils.core import setup

    old_path = os.getcwd()
    local_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    src_path = local_path

    os.chdir(local_path)
    sys.path.insert(0, local_path)

    # Run build
    old_path = os.getcwd()
    os.chdir(src_path)
    sys.path.insert(0, src_path)

    cwd = os.path.abspath(os.path.dirname(__file__))
    if not os.path.exists(os.path.join(cwd, 'PKG-INFO')):
        # Generate Cython sources, unless building from source release
        generate_cython(PACKAGE_NAME)

    try:
        setup(name=PACKAGE_NAME,
              author=AUTHOR,
              author_email=AUTHOR_EMAIL,
              url=URL,
              download_url=DOWNLOAD_URL,
              description=DESCRIPTION,
              long_description = LONG_DESCRIPTION,
              version=version(PACKAGE_NAME),
              license=LICENSE,
              configuration=configuration,
              keywords = ['time series','machine learning','bayesian statistics'],
              install_requires=['numpy', 'pandas', 'scipy', 'matplotlib', 'seaborn','numdifftools','patsy'])
    finally:
        del sys.path[0]
        os.chdir(old_path)

    return

if __name__ == '__main__':
    setup_package()


import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.sparse as sp
import scipy.stats as ss
from scipy.stats import multivariate_normal
import matplotlib.pyplot as plt
import seaborn as sns

from .. import arma
from .. import inference as ifr
from .. import distributions as dst
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
from .. import data_check as dc

from .kernels import *

class GPNARX(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** GAUSSIAN PROCESS NONLINEAR AUTOREGRESSIVE (GP-NARX) MODELS ****
    
    Parameters
    ----------
    data : pd.DataFrame or np.array
        Field to specify the time series data that will be used.
    
    ar : int
        Field to specify how many AR terms the model will have.
    
    kernel : kernel object
        For example, SquaredExponential() or ARD()
    
    integ : int (default : 0)
        Specifies how many time to difference the time series.
    
    target : str (pd.DataFrame) or int (np.array)
        Specifies which column name or array index to use. By default, first
        column/array will be selected as the dependent variable.

    sparse_gp : boolean 
        Whether to use sparse GP approximation with inducing points or not

    inducing_points : int
        How many inducing points to use if performing a sparse GP

    """

    def __init__(self, data, ar, kernel, integ=0, target=None, sparse_gp=False, inducing_points=6):

        # Initialize TSM object
        super(GPNARX,self).__init__('GPNARX')

        # Latent variables
        self.ar = ar
        self.integ = integ
        self.sparse_gp = sparse_gp

        self.max_lag = self.ar
        self.model_name = 'GPNARX(' + str(self.ar) + ')'
        self._z_hide = 0 # Whether to cutoff variance latent variables from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"     

        self.multivariate_model = False

        # Format the data
        self.data, self.data_name, self.is_pandas, self.index = dc.data_check(data,target)
        self.data_original = self.data.copy()

        # Difference data
        for order in range(self.integ):
            self.data = np.diff(self.data)
            self.data_name = "Differenced " + self.data_name
        self.index = self.index[self.integ:len(self.index)]

        # Apply normalization
        self.data_full = self.data.copy()       
        self.data = np.array(self.data_full[self.max_lag:self.data_full.shape[0]]) # adjust for lags
        self._norm_mean = np.mean(self.data)
        self._norm_std = np.std(self.data)  
        self.data = (self.data - self._norm_mean) / self._norm_std
        self.data_full = (self.data_full - self._norm_mean) / self._norm_std

        self.kernel = kernel
        self.kernel.X = self.X().T

        # Define latent variables
        self._create_latent_variables()

        if self.sparse_gp is True:
            self.neg_loglik = self.variational_neg_loglik
            self.inducing_point_no = inducing_points
            self.inducing_points = np.random.choice(self.X().shape[1], inducing_points, replace=False)
            self.sparse_X = self.X().T[self.inducing_points,:]
        else:
            self.neg_loglik = self.full_neg_loglik

    def _alpha(self, L):
        """ Covariance-derived term to construct expectations. See Rasmussen & Williams.
        
        Parameters
        ----------
        L : np.ndarray
            Cholesky triangular

        Returns
        ----------
        np.ndarray (alpha)
        """     
        return np.linalg.solve(np.transpose(L),np.linalg.solve(L,np.transpose(self.data)))

    def _low_rank_matrix(self, parm):
        """ Low rank (Nystrom) matrix approximation

        Parameters
        ----------
        parm : np.ndarray
            Contains transformed latent variables for kernels
        """        
        Km_inv = np.linalg.pinv(self.kernel.K_arbitrary_X(parm, self.sparse_X, self.sparse_X)+np.eye(self.inducing_point_no)*10**-8)
        Kxm = self.kernel.Kstar(parm, self.sparse_X)
        return(np.dot(np.dot(Kxm, Km_inv), Kxm.T))

    def _construct_predict(self, beta, h):    
        """ Creates h-step ahead forecasts for the Gaussian process
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for THE latent variables
        
        h: int
            How many steps ahead to forecast
        
        Returns
        ----------
        - predictions
        - variance of predictions
        """             

        # Refactor this entire code in future
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        Xstart = self.X().copy()
        Xstart = [i for i in Xstart]
        predictions = np.zeros(h)
        variances = np.zeros(h)

        for step in range(0,h):
            Xstar = []

            for lag in range(0,self.max_lag):
                if lag == 0:
                    if step == 0:
                        Xstar.append([self.data[-1]])
                        Xstart[0] = np.append(Xstart[0],self.data[-1])
                    else:
                        Xstar.append([predictions[step-1]])
                        Xstart[0] = np.append(Xstart[0],predictions[step-1])
                else:
                    Xstar.append([Xstart[lag-1][-2]])
                    Xstart[lag] = np.append(Xstart[lag],Xstart[lag-1][-2])

            Kstar = self.kernel.Kstar(parm, np.transpose(np.array(Xstar)))

            L = self._L(parm)
            
            if self.sparse_gp is True:
                alpha = self._alpha_cg(parm)   

            else:
                L = self._L(parm)
                alpha = self._alpha(L)   

            predictions[step] = np.dot(np.transpose(Kstar),alpha)
            variances[step] = self.kernel.Kstarstar(parm, np.transpose(np.array(Xstar))) 
            - np.dot(np.dot(np.transpose(self.kernel.Kstar(parm, np.transpose(np.array(Xstar)))),
                np.linalg.inv(self.kernel.K(parm) + np.identity(self.kernel.K(parm).shape[0])*parm[0])),
                self.kernel.Kstar(parm, np.transpose(np.array(Xstar)))) + parm[0] 

        return predictions, variances, predictions - 1.98*np.power(variances,0.5), predictions + 1.98*np.power(variances,0.5)

    def _create_latent_variables(self):
        """ Creates model latent variables

        Returns
        ----------
        None (changes model attributes)
        """

        # Create latent variables
        for no, i in enumerate(self.kernel.build_latent_variables()):
            self.latent_variables.add_z(i[0],i[1],i[2])
            self.latent_variables.z_list[no].start = i[3]

        self.z_no = len(self.kernel.build_latent_variables())

        # Use an ARIMA model to find starting point for the initial noise latent variable
        arma_start = arma.ARIMA(self.data,ar=self.ar,ma=0,integ=self.integ)
        x = arma_start.fit()
        arma_starting_values = arma_start.latent_variables.get_z_values()
        self.latent_variables.z_list[0].start = np.log(np.exp(np.power(arma_starting_values[-1],2)))

    def _L(self, parm):
        """ Creates cholesky decomposition of covariance matrix

        Parameters
        ----------
        parm : np.array
            Contains transformed latent variables

        Returns
        ----------
        The cholesky decomposition (L) of K
        """ 

        return np.linalg.cholesky(self.kernel.K(parm)) + np.identity(self.X().shape[1])*parm[0]

    def X(self):
        """ Creates design matrix of variables to use in GP regression
        
        Returns
        ----------
        The design matrix
        """     

        for i in range(0,self.ar):
            datapoint = self.data_full[(self.max_lag-i-1):-i-1]         
            if i == 0:
                X = datapoint
            else:
                X = np.vstack((X,datapoint))
        return X

    def expected_values(self, beta):
        """ Expected values of the function given the covariance matrix and hyperparameters
        
        Parameters
        ----------
        beta : np.ndarray
            Contains untransformed values for latent variables
        
        Returns
        ----------
        The expected values of the function
        """     

        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        L = self._L(parm)
        alpha = self._alpha(L)
        return np.dot(np.transpose(self.kernel.K(parm)), alpha)

    def variance_values(self, beta):
        """ Covariance matrix for the estimated function
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        Returns
        ----------
        Covariance matrix for the estimated function 
        """     
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        return self.kernel.K(parm) - np.dot(np.dot(np.transpose(self.kernel.K(parm)),np.linalg.inv(self.kernel.K(parm) + np.identity(self.kernel.K(parm).shape[0])*parm[0])),self.kernel.K(parm)) + parm[0]

    def variational_neg_loglik(self, beta):
        """ Lower bound variational approximation to log marginal loglikelihood
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        Returns
        ----------
        The variational lower bound
        """        
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        low_rank_matrix = self._low_rank_matrix(parm) + (parm[0]*np.identity(self.X().shape[1]))
        
        # apply Woodbury's matrix identity
        Kxm = self.kernel.Kstar(parm, self.sparse_X)
        Km = self.kernel.K_arbitrary_X(parm, self.sparse_X, self.sparse_X)+np.eye(self.inducing_point_no)*10**-8
        Km_inv = np.linalg.pinv(Km+np.identity(Km.shape[0])*(10**-8))
    
        first_term = np.linalg.pinv(parm[0]*np.identity(self.X().shape[1])+(10**-8)*np.identity(self.X().shape[1]))
        second_1 = np.dot(np.dot(np.dot(first_term, Kxm), (Km + np.dot(np.dot(Kxm.T,first_term),Kxm))),np.dot(Kxm.T,first_term))
        inverse_nystrom = first_term - second_1
        multivariate_normal_lik = -0.5*np.log(2.0*np.pi) -0.5*np.linalg.slogdet(low_rank_matrix)[1] - 0.5*(np.dot(np.dot(self.data.T,inverse_nystrom),self.data))

        return -( multivariate_normal_lik -(1.0/(2.0*parm[0]))*np.trace(self.kernel.K(parm)-low_rank_matrix))

    def full_neg_loglik(self, beta):
        """ Creates the negative log marginal likelihood of the model
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        Returns
        ----------
        The negative log marginal logliklihood of the model
        """             
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        L = self._L(parm)
        return -(-0.5*(np.dot(np.transpose(self.data),self._alpha(L))) - np.trace(np.log(L))- (self.data.shape[0]/2)*np.log(2*np.pi))

    def plot_fit(self, intervals=True, **kwargs):
        """ Plots the fit of the Gaussian process model to the data

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        intervals : Boolean
            Whether to plot uncertainty intervals or not
        
        Returns
        ----------
        None (plots the fit of the function)
        """

        figsize = kwargs.get('figsize',(10,7))

        date_index = self.index[self.max_lag:]
        expectation = self.expected_values(self.latent_variables.get_z_values())
        posterior = multivariate_normal(expectation, self.variance_values(self.latent_variables.get_z_values()))
        simulations = 500
        sim_vector = np.zeros([simulations,len(expectation)])

        for i in range(simulations):
            sim_vector[i] = posterior.rvs()

        error_bars = []
        for pre in range(5,100,5):
            error_bars.append([(np.percentile(i,pre)*self._norm_std + self._norm_mean) for i in sim_vector.transpose()] 
                - (expectation*self._norm_std + self._norm_mean))

        plt.figure(figsize=figsize) 

        plt.subplot(2, 2, 1)
        plt.title(self.data_name + " Raw")  
        plt.plot(date_index,self.data*self._norm_std + self._norm_mean,'k')

        plt.subplot(2, 2, 2)
        plt.title(self.data_name + " Raw and Expected") 
        plt.plot(date_index,self.data*self._norm_std + self._norm_mean,'k',alpha=0.2)
        plt.plot(date_index,self.expected_values(self.latent_variables.get_z_values())*self._norm_std + self._norm_mean,'b')

        plt.subplot(2, 2, 3)
        plt.title(self.data_name + " Raw and Expected (with intervals)")    

        if intervals == True:
            alpha =[0.15*i/float(100) for i in range(50,12,-2)]
            for count, pre in enumerate(error_bars):
                plt.fill_between(date_index, (expectation*self._norm_std + self._norm_mean)-pre, 
                    (expectation*self._norm_std + self._norm_mean)+pre,alpha=alpha[count])      

        plt.plot(date_index,self.data*self._norm_std + self._norm_mean,'k',alpha=0.2)
        plt.plot(date_index,self.expected_values(self.latent_variables.get_z_values())*self._norm_std + self._norm_mean,'b')

        plt.subplot(2, 2, 4)

        plt.title("Expected " + self.data_name + " (with intervals)")   

        if intervals == True:
            alpha =[0.15*i/float(100) for i in range(50,12,-2)]
            for count, pre in enumerate(error_bars):
                plt.fill_between(date_index, (expectation*self._norm_std + self._norm_mean)-pre, 
                    (expectation*self._norm_std + self._norm_mean)+pre,alpha=alpha[count])      

        plt.plot(date_index,self.expected_values(self.latent_variables.get_z_values())*self._norm_std + self._norm_mean,'b')

        plt.show()

    def plot_predict(self, h=5, past_values=20, intervals=True,**kwargs):
        """ Plots forecast with the estimated model
        
        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?
        
        past_values : int (default : 20)
            How many past observations to show on the forecast graph?
        
        intervals : Boolean
            Would you like to show 95% prediction intervals for the forecast?
        
        Returns
        ----------
        - Plot of the forecast
        - Error bars, forecasted_values, plot_values, plot_index
        """     

        figsize = kwargs.get('figsize',(10,7))

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:

            predictions, variance, lower, upper = self._construct_predict(self.latent_variables.get_z_values(),h) 
            full_predictions = np.append(self.data,predictions)
            full_lower = np.append(self.data,lower)
            full_upper = np.append(self.data,upper)
            date_index = self.shift_dates(h)

            # Plot values (how far to look back)
            plot_values = full_predictions[-h-past_values:]*self._norm_std + self._norm_mean
            plot_index = date_index[-h-past_values:]

            # Lower and upper intervals
            lower = np.append(full_predictions[-h-1],lower)
            upper = np.append(full_predictions[-h-1],upper)

            plt.figure(figsize=figsize)
            if intervals == True:
                plt.fill_between(date_index[-h-1:], 
                    lower*self._norm_std + self._norm_mean, 
                    upper*self._norm_std + self._norm_mean,
                    alpha=0.2)          
            
            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.data_name)
            plt.xlabel("Time")
            plt.ylabel(self.data_name)
            plt.show()

    def predict_is(self, h=5):
        """ Makes dynamic in-sample predictions with the estimated model
        
        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?
        
        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            x = GPNARX(ar=self.ar,kernel=self.kernel,integ=self.integ,
                sparse_gp=self.sparse_gp, inducing_points=self.inducing_point_no,
                data=self.data_original[:-h+t])
            if t == 0:
                x.fit(printer=False)
                predictions = x.predict(1)  
            else:
                x.fit(printer=False)
                predictions = pd.concat([predictions,x.predict(1)])

        predictions.rename(columns={0:self.data_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self, h=5, **kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)
        
        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast
        
        Returns
        ----------
        - Plot of the forecast against data 
        """     

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        date_index = self.index[-h:]
        predictions = self.predict_is(h)
        data = self.data[-h:]

        plt.plot(date_index,data*self._norm_std + self._norm_mean,label='Data')
        plt.plot(date_index,predictions,label='Predictions',c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def predict(self,h=5):
        """ Makes forecast with the estimated model
        
        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?
        
        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            predictions, _, _, _ = self._construct_predict(self.latent_variables.get_z_values(),h)    
            predictions = predictions*self._norm_std + self._norm_mean  
            date_index = self.shift_dates(h)
            result = pd.DataFrame(predictions)
            result.rename(columns={0:self.data_name}, inplace=True)
            result.index = date_index[-h:]

            return result


from .. import families as fam
from .. import tsm as tsm

from .dynlin import *
from .ndynlin import *

class DynamicGLM(tsm.TSM):
    """ Wrapper for dynamic GLM models

    Parameters
    ----------
    formula : str
        Patsy string specifying the regression

    data : pd.DataFrame or np.array
        Field to specify the time series data that will be used.

    family : 
        e.g. pf.Normal(0,1)
    """

    def __new__(cls, formula, data, family):
        if isinstance(family, fam.Normal):
            return DynReg(formula=formula, data=data)
        else:
            return NDynReg(formula=formula, data=data, family=family)

    def __init__(self, formula, data, family):
        pass




import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import scipy.stats as ss

from .stoch_optim import RMSProp, ADAM

from .bbvi_routines import alpha_recursion, log_p_posterior, mb_log_p_posterior


class BBVI(object):
    """
    Black Box Variational Inference
    
    Parameters
    ----------
    neg_posterior : function
        posterior function

    q : List
        list holding distribution objects
    
    sims : int
        Number of Monte Carlo sims for the gradient
    
    step : float
        Step size for RMSProp
    
    iterations: int
        How many iterations to run

    record_elbo : boolean
        Whether to record the ELBO at every iteration
    """

    def __init__(self, neg_posterior, q, sims, optimizer='RMSProp', iterations=1000, learning_rate=0.001, record_elbo=False):
        self.neg_posterior = neg_posterior
        self.q = q
        self.sims = sims
        self.iterations = iterations
        self.approx_param_no = np.array([i.param_no for i in self.q])
        self.optimizer = optimizer
        self.printer = True
        self.learning_rate = learning_rate
        self.record_elbo = record_elbo

    def change_parameters(self,params):
        """
        Utility function for changing the approximate distribution parameters
        """
        no_of_params = 0
        for core_param in range(len(self.q)):
            for approx_param in range(self.q[core_param].param_no):
                self.q[core_param].vi_change_param(approx_param, params[no_of_params])
                no_of_params += 1

    def create_normal_logq(self,z):
        """
        Create logq components for mean-field normal family (the entropy estimate)
        """
        means, scale = self.get_means_and_scales()
        return ss.norm.logpdf(z,loc=means,scale=scale).sum()

    def current_parameters(self):
        """
        Obtains an array with the current parameters
        """
        current = []
        for core_param in range(len(self.q)):
            for approx_param in range(self.q[core_param].param_no):
                current.append(self.q[core_param].vi_return_param(approx_param))
        return np.array(current)

    def cv_gradient(self,z):
        """
        The control variate augmented Monte Carlo gradient estimate
        """
        gradient = np.zeros(np.sum(self.approx_param_no))
        z_t = z.T      
        log_q = self.normal_log_q(z.T)
        log_p = self.log_p(z.T)
        grad_log_q = self.grad_log_q(z)
        gradient = grad_log_q*(log_p-log_q)

        alpha0 = alpha_recursion(np.zeros(np.sum(self.approx_param_no)), grad_log_q, gradient, np.sum(self.approx_param_no))           

        vectorized = gradient - ((alpha0/np.var(grad_log_q,axis=1))*grad_log_q.T).T

        return np.mean(vectorized,axis=1)

    def cv_gradient_initial(self,z):
        """
        The control variate augmented Monte Carlo gradient estimate
        """
        gradient = np.zeros(np.sum(self.approx_param_no))
        z_t = z.T      
        log_q = self.normal_log_q_initial(z.T)
        log_p = self.log_p(z.T)
        grad_log_q = self.grad_log_q(z)
        gradient = grad_log_q*(log_p-log_q)

        alpha0 = alpha_recursion(np.zeros(np.sum(self.approx_param_no)), grad_log_q, gradient, np.sum(self.approx_param_no))

        vectorized = gradient - ((alpha0/np.var(grad_log_q,axis=1))*grad_log_q.T).T

        return np.mean(vectorized,axis=1)

    def draw_normal(self):
        """
        Draw parameters from a mean-field normal family
        """
        means, scale = self.get_means_and_scales()
        return np.random.normal(means,scale,size=[self.sims,means.shape[0]]).T

    def draw_normal_initial(self):
        """
        Draw parameters from a mean-field normal family
        """
        means, scale = self.get_means_and_scales_from_q()
        return np.random.normal(means,scale,size=[self.sims,means.shape[0]]).T

    def draw_variables(self):
        """
        Draw parameters from the approximating distributions
        """        
        z = self.q[0].draw_variable_local(self.sims)
        for i in range(1,len(self.q)):
            z = np.vstack((z,self.q[i].draw_variable_local(self.sims)))
        return z

    def get_means_and_scales_from_q(self):
        """
        Gets the mean and scales for normal approximating parameters
        """
        means = np.zeros(len(self.q))
        scale = np.zeros(len(self.q))
        for i in range(len(self.q)):
            means[i] = self.q[i].mu0
            scale[i] = self.q[i].sigma0 
        return means, scale

    def get_means_and_scales(self):
        """
        Gets the mean and scales for normal approximating parameters
        """
        return self.optim.parameters[::2], np.exp(self.optim.parameters[1::2])

    def grad_log_q(self,z):
        """
        The gradients of the approximating distributions
        """        
        param_count = 0
        grad = np.zeros((np.sum(self.approx_param_no),self.sims))
        for core_param in range(len(self.q)):
            for approx_param in range(self.q[core_param].param_no):
                grad[param_count] = self.q[core_param].vi_score(z[core_param],approx_param)        
                param_count += 1
        return grad

    def log_p(self,z):
        """
        The unnormalized log posterior components (the quantity we want to approximate)
        """
        return log_p_posterior(z, self.neg_posterior)

    def normal_log_q(self,z):
        """
        The mean-field normal log posterior components (the quantity we want to approximate)
        """
        means, scale = self.get_means_and_scales()
        return (ss.norm.logpdf(z,loc=means,scale=scale)).sum(axis=1)

    def normal_log_q_initial(self,z):
        """
        The mean-field normal log posterior components (the quantity we want to approximate)
        """
        means, scale = self.get_means_and_scales_from_q()
        return (ss.norm.logpdf(z,loc=means,scale=scale)).sum(axis=1)

    def print_progress(self, i, current_params):
        """
        Prints the current ELBO at every decile of total iterations
        """
        for split in range(1,11):
            if i == (round(self.iterations/10*split)-1):
                post = -self.neg_posterior(current_params)
                approx = self.create_normal_logq(current_params)
                diff = post - approx
                print(str(split) + "0% done : ELBO is " + str(diff) + ", p(y,z) is " + str(post) + ", q(z) is " + str(approx))

    def get_elbo(self, current_params):
        """
        Obtains the ELBO for the current set of parameters
        """
        return -self.neg_posterior(current_params) - self.create_normal_logq(current_params)

    def run(self):
        """
        The core BBVI routine - draws Monte Carlo gradients and uses a stochastic optimizer.
        """

        # Initialization assumptions
        z = self.draw_normal_initial()
        gradient = self.cv_gradient_initial(z)
        gradient[np.isnan(gradient)] = 0
        variance = np.power(gradient, 2)       
        final_parameters = self.current_parameters()
        final_samples = 1

        # Create optimizer
        if self.optimizer == 'ADAM':
            self.optim = ADAM(final_parameters, variance, self.learning_rate, 0.9, 0.999)
        elif self.optimizer == 'RMSProp':
            self.optim = RMSProp(final_parameters, variance, self.learning_rate, 0.99)

        # Record elbo
        if self.record_elbo is True:
            elbo_records = np.zeros(self.iterations)
        else:
            elbo_records = None

        for i in range(self.iterations):
            x = self.draw_normal()
            gradient = self.cv_gradient(x)
            gradient[np.isnan(gradient)] = 0
            self.change_parameters(self.optim.update(gradient))

            if self.printer is True:
                self.print_progress(i, self.optim.parameters[::2])

            # Construct final parameters using final 10% of samples
            if i > self.iterations-round(self.iterations/10):
                final_samples += 1
                final_parameters = final_parameters+self.optim.parameters

            if self.record_elbo is True:
                elbo_records[i] = self.get_elbo(self.optim.parameters[::2])

        final_parameters = final_parameters/float(final_samples)
        self.change_parameters(final_parameters)
        final_means = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2==0])
        final_ses = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2!=0])
        if self.printer is True:
            print("")
            print("Final model ELBO is " + str(-self.neg_posterior(final_means)-self.create_normal_logq(final_means)))
        return self.q, final_means, final_ses, elbo_records

    def run_and_store(self):
        """
        The core BBVI routine - draws Monte Carlo gradients and uses a stochastic optimizer. 
        Stores rgw history of updates for the benefit of a pretty animation.
        """
        # Initialization assumptions
        z = self.draw_normal_initial()
        gradient = self.cv_gradient_initial(z)
        gradient[np.isnan(gradient)] = 0
        variance = np.power(gradient,2)       
        final_parameters = self.current_parameters()
        final_samples = 1

        # Create optimizer
        if self.optimizer == 'ADAM':
            self.optim = ADAM(final_parameters, variance, self.learning_rate, 0.9, 0.999)
        elif self.optimizer == 'RMSProp':
            self.optim = RMSProp(final_parameters, variance, self.learning_rate, 0.99)

        # Stored updates
        stored_means = np.zeros((self.iterations,len(final_parameters)/2))
        stored_predictive_likelihood = np.zeros(self.iterations)

        # Record elbo
        if self.record_elbo is True:
            elbo_records = np.zeros(self.iterations)
        else:
            elbo_records = None

        for i in range(self.iterations):
            gradient = self.cv_gradient(self.draw_normal())
            gradient[np.isnan(gradient)] = 0
            new_parameters = self.optim.update(gradient)
            self.change_parameters(new_parameters)

            stored_means[i] = self.optim.parameters[::2]
            stored_predictive_likelihood[i] = self.neg_posterior(stored_means[i])

            if self.printer is True:
                self.print_progress(i,self.optim.parameters[::2])

            # Construct final parameters using final 10% of samples
            if i > self.iterations-round(self.iterations/10):
                final_samples += 1
                final_parameters = final_parameters+self.optim.parameters

            if self.record_elbo is True:
                elbo_records[i] = self.get_elbo(self.optim.parameters[::2])

        final_parameters = final_parameters/float(final_samples)
        self.change_parameters(final_parameters)
        final_means = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2==0])
        final_ses = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2!=0])

        if self.printer is True:
            print("")
            print("Final model ELBO is " + str(-self.neg_posterior(final_means)-self.create_normal_logq(final_means)))
        return self.q, final_means, final_ses, stored_means, stored_predictive_likelihood, elbo_records

class CBBVI(BBVI):

    def __init__(self, neg_posterior, log_p_blanket, q, sims, optimizer='RMSProp',iterations=300000, 
        learning_rate=0.001, record_elbo=False):
        super(CBBVI, self).__init__(neg_posterior, q, sims, optimizer, iterations, learning_rate, record_elbo)
        self.log_p_blanket = log_p_blanket

    def log_p(self,z):
        """
        The unnormalized log posterior components (the quantity we want to approximate)
        RAO-BLACKWELLIZED!
        """      
        return np.array([self.log_p_blanket(i) for i in z])

    def normal_log_q(self,z):
        """
        The unnormalized log posterior components for mean-field normal family (the quantity we want to approximate)
        RAO-BLACKWELLIZED!
        """             
        means, scale = self.get_means_and_scales()
        return ss.norm.logpdf(z,loc=means,scale=scale)

    def normal_log_q_initial(self,z):
        """
        The unnormalized log posterior components for mean-field normal family (the quantity we want to approximate)
        RAO-BLACKWELLIZED!
        """             
        means, scale = self.get_means_and_scales_from_q()
        return ss.norm.logpdf(z,loc=means,scale=scale)

    def cv_gradient(self, z):
        """
        The control variate augmented Monte Carlo gradient estimate
        RAO-BLACKWELLIZED!
        """        
        z_t = np.transpose(z)
        log_q = self.normal_log_q(z_t)
        log_p = self.log_p(z_t)
        grad_log_q = self.grad_log_q(z)
        gradient = grad_log_q*np.repeat((log_p - log_q).T,2,axis=0)

        alpha0 = alpha_recursion(np.zeros(np.sum(self.approx_param_no)), grad_log_q, gradient, np.sum(self.approx_param_no))      
        
        vectorized = gradient - ((alpha0/np.var(grad_log_q,axis=1))*grad_log_q.T).T

        return np.mean(vectorized,axis=1)

    def cv_gradient_initial(self,z):
        """
        The control variate augmented Monte Carlo gradient estimate
        RAO-BLACKWELLIZED!
        """        
        z_t = np.transpose(z)
        log_q = self.normal_log_q_initial(z_t)
        log_p = self.log_p(z_t)
        grad_log_q = self.grad_log_q(z)
        gradient = grad_log_q*np.repeat((log_p - log_q).T,2,axis=0)

        alpha0 = alpha_recursion(np.zeros(np.sum(self.approx_param_no)), grad_log_q, gradient, np.sum(self.approx_param_no))

        vectorized = gradient - ((alpha0/np.var(grad_log_q,axis=1))*grad_log_q.T).T

        return np.mean(vectorized,axis=1)


class BBVIM(BBVI):
    """
    Black Box Variational Inference - minibatch
    
    Parameters
    ----------
    neg_posterior : function
        posterior function

    full_neg_posterior : function
        posterior function

    q : List
        list holding distribution objects

    sims : int
        Number of Monte Carlo sims for the gradient

    step : float
        Step size for RMSProp

    iterations: int
        How many iterations to run

    mini_batch : int
        Mini batch size
    """

    def __init__(self, neg_posterior, full_neg_posterior, q, sims, optimizer='RMSProp', 
        iterations=1000, learning_rate=0.001, mini_batch=2, record_elbo=False):
        self.neg_posterior = neg_posterior
        self.full_neg_posterior = full_neg_posterior
        self.q = q
        self.sims = sims
        self.iterations = iterations
        self.approx_param_no = np.array([i.param_no for i in self.q])
        self.optimizer = optimizer
        self.printer = True
        self.learning_rate = learning_rate
        self.mini_batch = mini_batch
        self.record_elbo = record_elbo

    def log_p(self,z):
        """
        The unnormalized log posterior components (the quantity we want to approximate)
        """
        return mb_log_p_posterior(z, self.neg_posterior, self.mini_batch)

    def get_elbo(self, current_params):
        """
        Obtains the ELBO for the current set of parameters
        """
        return -self.full_neg_posterior(current_params) - self.create_normal_logq(current_params)

    def print_progress(self, i, current_params):
        """
        Prints the current ELBO at every decile of total iterations
        """
        for split in range(1,11):
            if i == (round(self.iterations/10*split)-1):
                post = -self.full_neg_posterior(current_params)
                approx = self.create_normal_logq(current_params)
                diff = post - approx
                print(str(split) + "0% done : ELBO is " + str(diff) + ", p(y,z) is " + str(post) + ", q(z) is " + str(approx))

    def run(self):
        """
        The core BBVI routine - draws Monte Carlo gradients and uses a stochastic optimizer.
        """

        # Initialization assumptions
        z = self.draw_normal_initial()
        gradient = self.cv_gradient_initial(z)
        gradient[np.isnan(gradient)] = 0
        variance = np.power(gradient, 2)       
        final_parameters = self.current_parameters()
        final_samples = 1

        # Create optimizer
        if self.optimizer == 'ADAM':
            self.optim = ADAM(final_parameters, variance, self.learning_rate, 0.9, 0.999)
        elif self.optimizer == 'RMSProp':
            self.optim = RMSProp(final_parameters, variance, self.learning_rate, 0.99)

        # Record elbo
        if self.record_elbo is True:
            elbo_records = np.zeros(self.iterations)
        else:
            elbo_records = None

        for i in range(self.iterations):
            x = self.draw_normal()
            gradient = self.cv_gradient(x)
            gradient[np.isnan(gradient)] = 0
            self.change_parameters(self.optim.update(gradient))

            if self.printer is True:
                self.print_progress(i, self.optim.parameters[::2])

            # Construct final parameters using final 10% of samples
            if i > self.iterations-round(self.iterations/10):
                final_samples += 1
                final_parameters = final_parameters+self.optim.parameters

            if self.record_elbo is True:
                elbo_records[i] = self.get_elbo(self.optim.parameters[::2])

        final_parameters = final_parameters/float(final_samples)
        self.change_parameters(final_parameters)
        final_means = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2==0])
        final_ses = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2!=0])
        if self.printer is True:
            print("")
            print("Final model ELBO is " + str(-self.full_neg_posterior(final_means)-self.create_normal_logq(final_means)))
        return self.q, final_means, final_ses, elbo_records

    def run_and_store(self):
        """
        The core BBVI routine - draws Monte Carlo gradients and uses a stochastic optimizer. 
        Stores rgw history of updates for the benefit of a pretty animation.
        """
        # Initialization assumptions
        z = self.draw_normal_initial()
        gradient = self.cv_gradient_initial(z)
        gradient[np.isnan(gradient)] = 0
        variance = np.power(gradient,2)       
        final_parameters = self.current_parameters()
        final_samples = 1

        # Create optimizer
        if self.optimizer == 'ADAM':
            self.optim = ADAM(final_parameters, variance, self.learning_rate, 0.9, 0.999)
        elif self.optimizer == 'RMSProp':
            self.optim = RMSProp(final_parameters, variance, self.learning_rate, 0.99)

        # Stored updates
        stored_means = np.zeros((self.iterations,len(final_parameters)/2))
        stored_predictive_likelihood = np.zeros(self.iterations)

        # Record elbo
        if self.record_elbo is True:
            elbo_records = np.zeros(self.iterations)
        else:
            elbo_records = None

        for i in range(self.iterations):
            gradient = self.cv_gradient(self.draw_normal())
            gradient[np.isnan(gradient)] = 0
            new_parameters = self.optim.update(gradient)
            self.change_parameters(new_parameters)

            stored_means[i] = self.optim.parameters[::2]
            stored_predictive_likelihood[i] = self.neg_posterior(stored_means[i])

            if self.printer is True:
                self.print_progress(i,self.optim.parameters[::2])

            # Construct final parameters using final 10% of samples
            if i > self.iterations-round(self.iterations/10):
                final_samples += 1
                final_parameters = final_parameters+self.optim.parameters

            if self.record_elbo is True:
                elbo_records[i] = self.get_elbo(self.optim.parameters[::2])

        final_parameters = final_parameters/float(final_samples)
        self.change_parameters(final_parameters)
        final_means = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2==0])
        final_ses = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2!=0])

        if self.printer is True:
            print("")
            print("Final model ELBO is " + str(-self.full_neg_posterior(final_means)-self.create_normal_logq(final_means)))
        return self.q, final_means, final_ses, stored_means, stored_predictive_likelihood, elbo_records



__version__ = "0.4.10"

from . import __check_build

from .arma import *
from .var import *
from .ensembles import *
from .families import *
from .gas import *
from .garch import *
from .gpnarx import *
from .inference import *
from .ssm import *
from .covariances import *
from .output import *
from .tests import *
